// muesli.c -*- C -*-
/* Interface to multiple interpreters.
   Copyright (C) 2008, 2009 University of Limerick

   This file is part of Mueslint.

   Mueslint is free software: you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation, either version 3 of the License, or (at your
   option) any later version.

   Mueslint is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with Mueslint.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef _MUESLI_INT_
#define _MUESLI_INT_

#include "../config.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <sys/times.h>
#include <unistd.h>

#ifdef HAVE_LIBREADLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#include "muesli.h"

const char *muesli_version = PACKAGE_VERSION;

/***********************/
/* Evaluator interface */
/***********************/

static evaluator_interface *evaluator_interfaces = NULL;

evaluator_interface *
muesli_make_evaluator_interface(const char* new_name,
				const char *new_extension,
				int new_is_binary,
				void (*setup_func)(evaluator_interface *myself),
				void (*init_func)(evaluator_interface *myself),
				option_handler_t new_option_handler,
				struct option *new_getopt_options,
				void* params,
				void *data)
{
  evaluator_interface *interface = (evaluator_interface*)malloc(sizeof(evaluator_interface));

  char *tmp = (char*)malloc(strlen(new_name)+1);
  strcpy(tmp, new_name);
  interface->name = tmp;
  tmp = (char*)malloc(strlen(new_extension)+1);
  strcpy(tmp, new_extension);
  interface->next = NULL;
  interface->extension = tmp;
  interface->binary = new_is_binary;
  interface->initialized = 0;
  interface->init_evaluator = init_func;
  interface->specializer = NULL;
  interface->add_function = NULL;
  interface->handle_option = new_option_handler;
  interface->getopt_options = new_getopt_options;
  interface->repl = NULL;
  interface->array_separator = (char*)", ";

  interface->language_verbosity = 0;
  interface->flags = 0;

  interface->underlying_program = NULL;
  interface->underlying_startup_string = NULL;
  interface->underlying_prompt_string = NULL;
  interface->underlying_evaluator_pid = 0;

  interface->app_params = params;
  interface->app_data = data;

  interface->eval_clear = NULL;
  interface->eval_give = NULL;
  interface->eval_given = NULL;

  interface->version = "?";

  (setup_func)(interface);

  return interface;
}

char*
muesli_getName(evaluator_interface *interface)
{
  return interface->name;
}

int
muesli_getBinary(evaluator_interface *interface)
{
  return interface->binary;
}

void
muesli_initialize_evaluator(evaluator_interface *interface)
{
  if (!interface->initialized) {
    if (interface->language_verbosity > 0) {
      fprintf(stderr, "Muesli initializing %s\n", interface->name);
    }
    (interface->init_evaluator)(interface);
    muesli_app_specialize(interface->name, interface);
    interface->initialized = 1;
  }
}

/**********/
/* Naming */
/**********/

evaluator_interface *
muesli_get_evaluator_by_extension(const char *extension)
{
  evaluator_interface *evaluator;
  for (evaluator = evaluator_interfaces;
       evaluator != NULL;
       evaluator = evaluator->next) {
    if ((evaluator->extension != NULL)
	&& strcmp(evaluator->extension, extension) == 0) {

      // Make sure it's ready for use
      muesli_initialize_evaluator(evaluator);

      return evaluator;
    }
  }
  return NULL;
}

const char *
muesli_get_language_by_extension(const char *extension)
{
  evaluator_interface *evaluator = muesli_get_evaluator_by_extension(extension);

  return (evaluator != NULL) ? evaluator->name : NULL;
}

int
muesli_evaluator_is_defined(const char *evaluator_name)
{
  evaluator_interface *evaluator;

  for (evaluator = evaluator_interfaces;
       evaluator != NULL;
       evaluator = evaluator->next) {
    if ((evaluator->extension != NULL)
	&& strcmp(evaluator->name, evaluator_name) == 0) {
      return 1;
    }
  }
  return 0;
}

/* This finds and initializes an evaluator.  As long as you always use
   this function to find yourself an evaluator, you should always get
   an initialized one.
 */
evaluator_interface *
muesli_get_named_evaluator(const char *evaluator_name, int exit_if_not_found)
{
  evaluator_interface *evaluator;

  for (evaluator = evaluator_interfaces;
       evaluator != NULL;
       evaluator = evaluator->next) {
    if ((evaluator->name != NULL)
	&& strcmp(evaluator->name, evaluator_name) == 0) {
      break;
    }
  }

  if (evaluator == NULL) {
    if (exit_if_not_found) {
      fprintf(stderr, "Evaluator \"%s\" not found\n", evaluator_name);
      exit(1);
    } else {
    return NULL;
    }
  }

  // Make sure it's ready for use
  muesli_initialize_evaluator(evaluator);

  return evaluator;
}

char *
muesli_result_type_name(result_type_t enumerated)
{
  switch (enumerated) {
  case no_result: return (char*)"void";
  case number_result: return (char*)"number";
  case string_result: return (char*)"string";
  case boolean_result: return (char*)"boolean";
  case error_result: return (char*)"error";
  default: return (char*)"unknown";
  }
}

result_type_t
muesli_eval_in_language(const char *language,
			const char *fragment,
			unsigned int fragment_length,
			eval_result_t *result_p,
			int transient)
{
  evaluator_interface *evaluator_interface = muesli_get_named_evaluator(language, 0);

  if (evaluator_interface == NULL) {
    fprintf(stderr, "No such evaluator: %s\n", language);
    return no_result;
  }

  return (evaluator_interface->eval_string)(evaluator_interface,
					    fragment,
					    fragment_length,
					    result_p,
					    transient);
}

int
muesli_load_file(const char *filename)
{
  const char *extension = strrchr(filename, '.');
  if (extension == NULL) {
    fprintf(stderr,
	    "Could not find extension of \"%s\" -- I do not know how to load this file\n",
	    filename);
    return 0;
  }
  extension++;
  if (*extension == '\0') {
    fprintf(stderr,
	    "Empty extension in \"%s\" -- I do not know how to load this file\n",
	    filename);
    return 0;
  }

  evaluator_interface *interface = muesli_get_evaluator_by_extension(extension);

  if (interface == NULL) {
    fprintf(stderr,
	    "I could not work out how to load \"%s\" from the extension \"%s\"\n",
	    filename, extension);
    return 0;
  }

  (interface->load_file)(interface, filename);
  return 1;
}

void
muesli_print_interpreter_names(FILE *stream,
			       const char *format,
			       const char *selected_name,
			       const char *selected_flag)
{
  evaluator_interface *eval;

  for (eval = evaluator_interfaces;
       eval != NULL;
       eval = eval->next) {
    fprintf(stream, format, eval->name,
	   (((selected_name != NULL)
	     && (strcmp(eval->name, selected_name) == 0))
	    ? (selected_flag ? selected_flag : " (selected)") : ""),
	   eval->version);
  }
}

/*******************/
/* Evaluator setup */
/*******************/

evaluator_interface *no_evaluator = NULL; // might put in a novelty one to print a message

evaluator_interface *scripting_evaluator = NULL;

evaluator_interface *evaluator_zero = NULL;
evaluator_interface *evaluator_siod = NULL;
evaluator_interface *evaluator_lua = NULL;
evaluator_interface *evaluator_external_exec = NULL;
evaluator_interface *evaluator_external_pipe = NULL;
evaluator_interface *evaluator_custom = NULL;
evaluator_interface *evaluator_perl = NULL;
evaluator_interface *evaluator_tcl = NULL;
evaluator_interface *evaluator_python = NULL;
evaluator_interface *evaluator_tcc = NULL;
evaluator_interface *evaluator_octave = NULL;
evaluator_interface *evaluator_slang = NULL;
evaluator_interface *evaluator_ruby = NULL;
evaluator_interface *evaluator_prolog = NULL;
evaluator_interface *evaluator_parrot = NULL;
evaluator_interface *evaluator_jvm = NULL;
evaluator_interface *evaluator_machine = NULL;
evaluator_interface *evaluator_stack_code = NULL;

/* You should use set_evaluator_app_specializer(string,
   specializer_function) to set a function to add your extensions.
   This is called when the interpreter is initialized, which happens
   when you call get_named_evaluator, so you must call
   set_evaluator_app_specializer before then.
*/
void
muesli_set_evaluator_app_specializer(char *evaluator_name,
			      evaluator_specializer_t specializer)
{
  evaluator_interface *evaluator;

  for (evaluator = evaluator_interfaces;
       evaluator != NULL;
       evaluator = evaluator->next) {
    if ((evaluator->name != NULL)
	&& strcmp(evaluator->name, evaluator_name) == 0) {
      break;
    }
  }

  if (evaluator != NULL) {
    evaluator->specializer = specializer;
  }
}

// This is an internal function; we call it for you.
// set_evaluator_app_specializer is the one you should call to set
// this up.
void
muesli_app_specialize(char *name,
	       evaluator_interface *interface)
{
  if (interface->specializer != NULL) {
    (interface->specializer)(interface);
  }
}

// A wrapper for new evaluator_interface, that registers the name in the map evaluators_by_name
evaluator_interface *
muesli_define_evaluator(const char *new_name,
			const char *new_extension,
			int new_is_binary,
			void (*new_setup)(evaluator_interface *myself),
			void (*new_init)(evaluator_interface *myself),
			option_handler_t new_option_handler,
			struct option *new_getopt_options,
			void *new_params,
			void *new_data)
{
  evaluator_interface *interface = muesli_make_evaluator_interface(new_name, new_extension, new_is_binary,
								   new_setup, new_init,
								   new_option_handler, new_getopt_options,
								   new_params, new_data);

  interface->next = evaluator_interfaces;
  evaluator_interfaces = interface;

#if 0
  fprintf(stderr, "  Added evaluator %p(%s) at %p, next %p\n",
	  interface,
	  interface->name,
	  evaluator_interfaces,
	  evaluator_interfaces->next  );
#endif

  if (new_is_binary
      && (interface->eval_0 == NULL)) {
    fprintf(stderr, "Warning: The evaluator \"%s\" claims to be binary, but cannot be\n         used as a binary fitness evaluator, as it lacks an eval_0 function\n", new_name);
  }

  return interface;
}

void
muesli_register_evaluators(void *new_app_params,
			   void *new_app_data,
			   option_handler_t params_handler,
			   struct option *getopt_options)
{
  // These create the interfaces, and give them the information
  // they'll need when the interpreters initialize.  These do not
  // initialize the interpreters; that happens later.
  evaluator_external_exec = muesli_define_evaluator("exec", "sh", 0,
						    external_exec_setup, exec_evaluator_init,
						    params_handler, getopt_options,
						    new_app_params, new_app_data);
  evaluator_external_pipe = muesli_define_evaluator("pipe", "", 0,
						    external_pipe_setup, pipe_evaluator_init,
						    params_handler, getopt_options,
						    new_app_params, new_app_data);
  evaluator_custom = muesli_define_evaluator("custom", "cus", 1,
					     custom_setup, custom_evaluator_init,
					     params_handler, getopt_options,
					     new_app_params, new_app_data);
  evaluator_zero = muesli_define_evaluator("zero", "zero", 1,
					   zero_setup, zero_evaluator_init,
					   params_handler, getopt_options,
					   new_app_params, new_app_data);
#if defined(HAVE_LIBSIOD) || defined(HAVE_LIBULSIOD)
  evaluator_siod = muesli_define_evaluator("siod", "scm", 0,
					   siod_setup, siod_evaluator_init,
					   params_handler, getopt_options,
					   new_app_params, new_app_data);
#endif
#ifdef HAVE_LIBLUA
  evaluator_lua = muesli_define_evaluator("lua", "lua", 0,
					  lua_setup, lua_evaluator_init,
					  params_handler, getopt_options,
					  new_app_params, new_app_data);
#endif
#ifdef HAVE_LIBPERL
  evaluator_perl = muesli_define_evaluator("perl", "pl", 0,
					   perl_setup, perl_evaluator_init,
					   params_handler, getopt_options,
					   new_app_params, new_app_data);
#endif
#ifdef HAVE_LIBTCL
  evaluator_tcl = muesli_define_evaluator("tcl", "tcl", 0,
					  tcl_setup, tcl_evaluator_init,
					  params_handler, getopt_options,
					  new_app_params, new_app_data);
#endif
#ifdef HAVE_LIBPYTHON
  evaluator_python = muesli_define_evaluator("python", "python", 0,
					     python_setup, python_evaluator_init,
					     params_handler, getopt_options,
					     new_app_params, new_app_data);
#endif
#ifdef HAVE_LIBTCC
  evaluator_tcc = muesli_define_evaluator("tcc", "c", 0,
					  tcc_setup, tcc_evaluator_init,
					  params_handler, getopt_options,
					  new_app_params, new_app_data);
#endif
#ifdef HAVE_LIBOCTINTERP
  evaluator_octave = muesli_define_evaluator("octave", "oct", 0,
					     octave_setup, octave_evaluator_init,
					     params_handler, getopt_options,
					     new_app_params, new_app_data);
#endif
#ifdef HAVE_LIBSLANG
  evaluator_slang = muesli_define_evaluator("slang", "sl", 0,
					    slang_setup, slang_evaluator_init,
					    params_handler, getopt_options,
					    new_app_params, new_app_data);
#endif
#ifdef HAVE_LIBRUBY
  evaluator_ruby = muesli_define_evaluator("ruby", "ruby", 0,
					   ruby_setup, ruby_evaluator_init,
					   params_handler, getopt_options,
					   new_app_params, new_app_data);
#endif
#ifdef HAVE_LIBPROLOG
  evaluator_prolog = muesli_define_evaluator("prolog", "pro", 0,
					     prolog_setup, prolog_evaluator_init,
					     params_handler, getopt_options,
					     new_app_params, new_app_data);
#endif
#ifdef HAVE_LIBPARROT
  evaluator_parrot = muesli_define_evaluator("parrot", "peg", 1,
					     parrot_setup, parrot_evaluator_init,
					     params_handler, getopt_options,
					     new_app_params, new_app_data);
#endif
#ifdef HAVE_LIBJVM
  evaluator_jvm = muesli_define_evaluator("jvm", "class", 1,
					  jvm_setup, jvm_evaluator_init,
					  params_handler, getopt_options,
					  new_app_params, new_app_data);
#endif
  evaluator_machine = muesli_define_evaluator("machine", "o", 1,
					      machine_code_setup, machine_code_evaluator_init,
					      params_handler, getopt_options,
					      new_app_params, new_app_data);
  evaluator_stack_code = muesli_define_evaluator("stack-code", "sc", 1,
						 stack_code_setup, stack_bytecode_evaluator_init,
						 params_handler, getopt_options,
						 new_app_params, new_app_data);
}

/**************************/
/* Interactive evaluation */
/**************************/

#define OUR_OWN_REPL 1

void
muesli_evaluator_repl(const char *evaluator_name)
{
  evaluator_interface *evaluator =
    muesli_get_named_evaluator(evaluator_name, 0);

  if (evaluator == NULL) {
    fprintf(stderr, "Could not find an evaluator called %s\n", evaluator_name);
  } else if (OUR_OWN_REPL || evaluator->repl == NULL) {
    int running = 1;
    char *prompt = (char*)malloc(strlen(evaluator->name) + 2);
    sprintf(prompt, "%s>", evaluator->name);
    fprintf(stderr, "Using makeshift Read-Eval-Print loop for %s\n", evaluator->name);
    fprintf(stderr, "Use quit to go back to application run or exit to leave application\n");
    do {
      char *input;
#ifdef HAVE_LIBREADLINE
      input = readline(prompt);
#else
      char input_buffer[CLI_BUF_SIZE];
      puts(prompt); fflush(stdout);
      fgets(input_buffer, CLI_BUF_SIZE, stdin);
      input = input_buffer;
#endif
      if (strcmp(input, "quit") == 0) {
	running = 0;
      } else if (strcmp(input, "exit") == 0) {
	exit(0);
      } else {
	eval_result_t result;
	result_type_t result_type = (evaluator->eval_string)(evaluator,
							     input,
							     strlen(input),
							     &result,
							     0);
	switch (result_type) {
	case no_result:
	  printf("no result (or nothing printable)\n");
	  break;
	case number_result:
	  printf("%f\n", result.as_number);
	  break;
	case string_result:
	  printf("\"%s\"\n", result.as_string);
	  break;
	case boolean_result:
	  printf("%s\n", result.as_bool ? "true" : "false");
	  break;
	case error_result:
	  printf("error\n");
	  break;
	}
      }
    } while (running);
    free(prompt);
  } else {
    printf("Calling Read-Eval-Print loop of %s\n", evaluator->name);
    evaluator->repl(evaluator);
  }
}

/*************************************/
/* Utility for shell/pipe evaluators */
/*************************************/

void
muesli_split_args(const char *string, char ***new_argv, int *new_argc)
{
  unsigned int n_parts = 1;
  const char *p = string;
  char c;
  int in_quotes = 0;
  int escaped = 0;
  char *scratch = (char*)malloc(1+strlen(string));
  char *q = scratch;

  // count the possible breaks between arguments -- possibly an
  // overestimate as some may be quoted, but that doesn't matter
  while ((c = *p++) != '\0') {
    switch (c) {
    case ' ': case '\t': case '\n': case '\r': n_parts++; break;
    default: break;
    }
  }

  *new_argv = (char**)malloc(n_parts * sizeof(char**));
  *new_argc = 0;
  (*new_argv)[0] = q;

  p = string;

  while ((c = *p++) != '\0') {
    if (escaped) {
      switch (c) {
      case 'n': *q++ = '\n'; break;
      case 'r': *q++ = '\r'; break;
      case 't': *q++ = '\t'; break;
      case '\\': *q++ = '\\'; break;
      default:  *q++ = c; break;
      }
      escaped = 0;
    } else {
      if (in_quotes) {
	switch (c) {
	case '\\': escaped = 1; break;
	case '"': in_quotes = 0; break;
	default: *q++ = c; break;
	}
      } else {
	switch (c) {
	case '\\': escaped = 1; break;
	case ' ': case '\t': case '\n': case '\r':
	  *q++ = '\0';
	  (*new_argv)[(*new_argc)++] = q;
	  break;
	case '"': in_quotes = 1; break;
	default: *q++ = c; break;
	}
      }
    }
  }
  (*new_argv)[(*new_argc)++] = NULL;
}

/*******************/
/* Option handling */
/*******************/

char
muesli_find_option_letter(struct option *getopt_options, const char *option_name)
{
  struct option *opt;

  for (opt = getopt_options; opt->name != NULL; opt++) {
    if (strcmp(option_name, opt->name) == 0) {
      return opt->val;
    }
  }
  return (char)(-1);
}

int
muesli_set_parameter(evaluator_interface *interface,
		     struct option *options,
		     char* option,
		     char *value,
		     const char *language)
{
  char option_code = muesli_find_option_letter(options, option);
  option_handler_t handler = interface->handle_option;

  // fprintf(stderr, "option_code=%c (%d)\n", option_code, option_code);

  if (option_code != -1) {
    result_type_t result_type;
    eval_result_t result;
    (handler)(interface->app_params, option_code,
	      value, 0.0, 1,
	      &result_type, &result, language);
    return 1;
  } else {
    return 0;
  }
}

static char num_buf[64];

const char*
muesli_get_parameter(evaluator_interface *interface,
		     struct option *options,
		     char* option)
{
  option_handler_t handler = interface->handle_option;
  char option_code = muesli_find_option_letter(options, option);
  if (option_code != -1) {
    result_type_t result_type;
    eval_result_t result;

    (handler)(interface->app_params,
	      option_code,	// option
	      NULL, 0.0,	// value
	      0,	// set
	      &result_type, &result,
	      NULL);

    switch (result_type) {
    case string_result:
      return result.as_string;
    case number_result:
      sprintf(num_buf, "%f", result.as_number);
      return num_buf;
    case boolean_result:
      if (result.as_int) {
	return "true";
      } else {
	return "false";
      }
    default:
      return NULL;
    }
  } else {
    return NULL;
  }
}

/**************/
/* Self-tests */
/**************/

static int
show_self_test_results(const char *label,
		       evaluator_interface *evaluator,
		       float reference_value,
		       result_type_t result_type,
		       eval_result_t string_eval_result)
{
  fprintf(stderr, "Test of %s evaluator:\n", label);
  if (evaluator == NULL) {
    fprintf(stderr, "  Not present\n");
    return 1;
  } else {
    switch (result_type) {
    case no_result:
      fprintf(stderr, "The evaluator returned no result\n");
      break;
    case number_result:
      fprintf(stderr, "  Number result is %f (%s)\n",
	      string_eval_result.as_number,
	      (string_eval_result.as_number == reference_value)
	      ? "right" : "wrong");
      break;
    case string_result:
      fprintf(stderr, "  String result is %s\n", string_eval_result.as_string);
      break;
    case boolean_result:
      fprintf(stderr, "  Boolean result is %s\n", string_eval_result.as_int ? "true" : "false");
      break;
    default:
      break;
    }
    return ((number_result == reference_value)
	    && (result_type == number_result));
  }
}

typedef float (*mc_test_function_t)(float, float, float);

typedef struct evaluator_test {
  char *name;
  char *test_string;
  int reference;
  int verbosity;
} evaluator_test;

int
test_evaluator(evaluator_test *test, int seconds, float *evals_per_second_p, float seconds_per_base_loop)
{
  char *name = test->name;
  char *test_string = test->test_string;
  int reference = test->reference;
  int verbosity = test->verbosity;
  evaluator_interface *evaluator = muesli_get_named_evaluator(name, 0);

  result_type_t result_type = no_result;
  eval_result_t result;
  result.as_number = 0.0;

  if (evaluator != NULL) {
    if (seconds > 0) {
      int ticks_per_sec = sysconf(_SC_CLK_TCK);

      printf("Timing %s on \"%s\" for %d seconds\n", name, test_string, seconds);

#if 1
      // You won't often want this one enabled, I'd wager; but it can
      // be useful for tracing gradually leaking resources, etc:
      evaluator->language_verbosity = verbosity;
#endif
      struct  tms eval_time;
       times(&eval_time);
      clock_t start_time = eval_time.tms_utime + eval_time.tms_stime + eval_time.tms_cutime + eval_time.tms_cstime;
      clock_t now_time = start_time;
      clock_t end_time = start_time + seconds * ticks_per_sec;
      int count = 0;
      while (now_time < end_time) {
	result_type = (evaluator->eval_string)(evaluator,
					       test_string,
					       strlen(test_string),
					       &result, 0);
	count++;
	times(&eval_time);
	now_time = eval_time.tms_utime + eval_time.tms_stime + eval_time.tms_cutime + eval_time.tms_cstime;
	if (result_type == no_result) {
	  printf("Problem in timing \"%s\"; it failed to return a result\n", name);
	  break;
	}
      }
      float evals_per_second = ((float)count) / (float)seconds;
      float seconds_per_eval = 1.0 / evals_per_second;
      float adjusted_seconds_per_eval = seconds_per_eval - seconds_per_base_loop;
      float adjusted_evals_per_second = 1.0 / adjusted_seconds_per_eval;

      printf("%s: %7.0f per second (raw: %7.0f)\n", name, adjusted_evals_per_second, evals_per_second);
      if (evals_per_second_p) {
	*evals_per_second_p = evals_per_second;
      }
    } else {
      printf("Testing %s on \"%s\" at verbosity %d\n", name, test_string, verbosity);
      evaluator->language_verbosity = verbosity;
      result_type = (evaluator->eval_string)(evaluator,
					     test_string,
					     strlen(test_string),
					     &result, 0);
    }
  } else {
    printf("Did not find an evaluator called \"%s\"\n", name);
  }
  return show_self_test_results(name, evaluator, reference, result_type, result);
}

evaluator_test evaluator_zero_test = { (char*)"zero", (char*)"", 0, 0};

evaluator_test evaluator_tests[] = {
#if 0
  {(char*)"exec", (char*)"dc -e \"4 3 * 2 + p q\"", 14, 0},
#endif
#if 0
  {(char*)"pipe", (char*)"", 14, 0},
  {(char*)"custom", (char*)"", 14, 0},
#endif
#if defined(HAVE_LIBSIOD) || defined(HAVE_LIBULSIOD)
  {(char*)"siod", (char*)"(+ (* 4 3) 2)", 14, 0},
#endif
#ifdef HAVE_LIBPERL
  {(char*)"perl", (char*)"return (4 * 3 + 2)", 14, 0},
#endif
#ifdef HAVE_LIBLUA
  {(char*)"lua", (char*)"return (4 * 3 + 2)", 14, 0},
#endif
#ifdef HAVE_LIB_SLANG
  {(char*)"s-lang", (char*)"return (4 * 3 + 2)", 14, 0},
#endif
#ifdef HAVE_LIBPYTHON
  {(char*)"python", (char*)"return (4 * 3 + 2)", 14, 0}, // todo: what is the syntax of this?
#endif
#ifdef HAVE_LIBTCL
  {(char*)"tcl", (char*)"expr (4 * 3 + 2)", 14, 0},
#endif
#ifdef HAVE_LIBTCC0
  {(char*)"tcc", (char*)"expr (4 * 3 + 2)", 14, 0},
#endif
#ifdef HAVE_LIBOCTAVE
  {(char*)"octave", (char*)"expr (4 * 3 + 2)", 14, 0},
#endif
#ifdef HAVE_LIBSLANG
  {(char*)"slang", (char*)"expr (4 * 3 + 2)", 14, 0},
#endif
#ifdef HAVE_LIBPARROT
  {(char*)"parrot", (char*)"expr (4 * 3 + 2)", 14, 0},
#endif
#ifdef HAVE_LIBJVM
  {(char*)"jvm", (char*)"expr (4 * 3 + 2)", 14, 0},
#endif
  {(char*)"stack-code", (char*)"cde*+", 14, 2},
  {NULL, NULL, 0}
};

int
muesli_test_evaluators(int seconds)
{
#if 0
  const char* test_string ="(load \"eval-test.scm\" nil t\n(run-tests eval-tests)\n";
  muesli_eval_in_language("siod", test_string, strlen(test_string), NULL);
#else
  int all_ok = 1;
  int i;
  result_type_t result_type = no_result;
  eval_result_t result;

  evaluator_interface *evaluator;

  // verbosity = verbose;

  float base_loops_per_second;
  float seconds_per_base_loop;

  test_evaluator(&evaluator_zero_test, seconds, &base_loops_per_second, 0.0);

  seconds_per_base_loop = 1.0 / base_loops_per_second;

  for (i = 0; evaluator_tests[i].name != NULL; i++) {
    test_evaluator(&evaluator_tests[i], seconds, NULL, seconds_per_base_loop);
  }

#if 0
  evaluator = get_named_evaluator("stack-code");

  if (evaluator != NULL) {
    int all_fns_present = 1;
    if (evaluator->eval_clear == NULL) {
      fprintf(stderr, "No eval_clear defined for stack-code\n");
      all_fns_present = 0;
    }
    if (evaluator->eval_give == NULL) {
      fprintf(stderr, "No eval_give defined for stack-code\n");
      all_fns_present = 0;
    }
    if (evaluator->eval_given == NULL) {
      fprintf(stderr, "No eval_given defined for stack-code\n");
      all_fns_present = 0;
    }
    if (all_fns_present) {

      evaluator->setLanguageVerbosity(2);

      evaluator->eval_clear(evaluator);
      evaluator->eval_give(evaluator, 2.0);
      evaluator->eval_give(evaluator, 3.0);
      evaluator->eval_give(evaluator, 4.0);
      result.as_number = evaluator->eval_given(evaluator, "*+", 2);
      result_type = number_result;
    } else {
      result_type = no_result;
    }

  }
  all_ok &= show_self_test_results("stack-bytecode", evaluator, 14, result_type, result);
#endif
  evaluator = muesli_get_named_evaluator("machine", 0);
  if (evaluator != NULL) {
    /*
      float
      test_fn(float a, float b, float c)
      {
        return a * b + c;
      }

      test_fn:
      55       		pushl	%ebp
      89E5     		movl	%esp, %ebp
      D94508   		flds	8(%ebp)
      D84D0C   		fmuls	12(%ebp)
      D84510   		fadds	16(%ebp)
      5D       		popl	%ebp
      C3       		ret
    */
    char machine_eval_test_string[] = {
      0x55,			// pushl %ebp
      0x89, 0xE5,		// movl	%esp, %ebp
      0xD9, 0x45, 0x08,		// flds	8(%ebp)
      0xD8, 0x4D, 0x0C,		// fmuls 12(%ebp)
      0xD8, 0x45, 0x10,		// fadds 16(%ebp)
      0x5D,			// popl	%ebp
      0xC3			// ret
    };
#if 1
    result.as_number = ((mc_test_function_t)(machine_eval_test_string))(3.0, 4.0, 2.0);
#else
    result_type = (evaluator->eval_string)(evaluator,
					   machine_eval_test_string,
					   sizeof(machine_eval_test_string),
					   &result, 0);
#endif
  }
  all_ok &= show_self_test_results("machine", evaluator, 14, result_type, result);

  printf(all_ok ? "All tests passed\n" : "At least one test failed\n");
#endif
  return all_ok;
}

#endif
